@@ -5,9 +5,9 @@ import time
 import traceback
 
 import redis
-
+from collections import OrderedDict
 from gevent.event import Event
-from suanpan import error, utils
+from suanpan import error, utils, g
 from suanpan.log import logger
 
 MQ_DELIVER_KEY = "_suanpan_mq_deliver"
@@ -15,21 +15,23 @@ MQ_DELIVER_KEY = "_suanpan_mq_deliver"
 
 class MQ(object):
     def __init__(
-        self,  # pylint: disable=unused-argument
-        redisHost="localhost",
-        redisPort=6379,
-        redisRealtime=False,
-        redisKeepalive=True,
-        # redisKeepaliveIDLE=120,
-        # redisKeepaliveCNT=2,
-        # redisKeepaliveINTVL=30,
-        redisSocketConnectTimeout=1,
-        redisUnixSocketPath=None,
-        options=None,
-        client=None,
-        **kwargs,
+            self,  # pylint: disable=unused-argument
+            redisHost="localhost",
+            redisPort=6379,
+            redisRealtime=False,
+            redisKeepalive=True,
+            # redisKeepaliveIDLE=120,
+            # redisKeepaliveCNT=2,
+            # redisKeepaliveINTVL=30,
+            redisSocketConnectTimeout=1,
+            redisUnixSocketPath=None,
+            options=None,
+            client=None,
+            __p2p_stream=False,
+            **kwargs,
     ):
         self.options = options or {}
+        self._p2p = __p2p_stream or kwargs.get('__p2p_stream', False)
         self.options.update(
             host=redisHost,
             port=redisPort,
@@ -67,6 +69,9 @@ class MQ(object):
         self._stop_event = Event()
         self._stop_event.clear()
 
+        if self._p2p:
+            self._getP2PConnections()
+
     def close(self):
         self._stop_event.set()
         self._stop_event.is_set()
@@ -115,7 +120,7 @@ class MQ(object):
         }
 
     def createQueue(
-        self, name, group="default", consumeID="0", force=False, existOk=False
+            self, name, group="default", consumeID="0", force=False, existOk=False
     ):
         logger.debug(f"Create Queue {name}-{group}")
         if force:
@@ -146,21 +151,48 @@ class MQ(object):
     #         return False
 
     def sendMessage(
-        self, queue, data, messageID="*", maxlen=1000, trimImmediately=False
+            self, queue, data, messageID="*", maxlen=1000, trimImmediately=False
     ):
-        return self.client.xadd(
-            queue, data, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
-        )
+        if self._p2p:
+            for target, data4in in self._sendP2PMessages(data):
+                target_queue = f'{queue}-{target}'
+                logger.debug(f'send p2p message to {target}: {data4in}')
+                self.client.xadd(
+                    target_queue, data4in, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
+                )
+        else:
+            return self.client.xadd(
+                queue, data, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
+            )
+
+    def _getP2PConnections(self):
+        from suanpan.graph import graph
+
+        self._outputConnections = OrderedDict()
+        node = graph.getNodeById(g.nodeId)
+        for port in node.metadata['def'].ports:
+            if port.ioType == 'out':
+                port_id = port.uuid
+                connections = graph.getOutputConnections(g.nodeId, port_id)
+                self._outputConnections[port_id] = connections
+
+    def _sendP2PMessages(self, data):
+        for port_id, connections in self._outputConnections.items():
+            for tgt in connections:
+                d = data.get(port_id)
+                if d:
+                    data4in = {"id": data['request_id'], tgt.port: d, "extra": data["extra"]}
+                    yield tgt.process, data4in
 
     def recvMessages(
-        self,
-        queue,
-        group="default",
-        consumer="unknown",
-        noack=False,
-        block=False,
-        count=1,
-        consumeID=">",
+            self,
+            queue,
+            group="default",
+            consumer="unknown",
+            noack=False,
+            block=False,
+            count=1,
+            consumeID=">",
     ):
         block = None if not block else 0 if block is True else block
         messages = self.client.xreadgroup(
@@ -180,17 +212,17 @@ class MQ(object):
         return [message for message in messages if message["data"]]
 
     def subscribeQueue(
-        self,
-        queue,
-        group="default",
-        consumer="unknown",
-        noack=False,
-        block=True,
-        count=1,
-        consumeID=">",
-        delay=0,
-        errDelay=1,
-        errCallback=logger.error,
+            self,
+            queue,
+            group="default",
+            consumer="unknown",
+            noack=False,
+            block=True,
+            count=1,
+            consumeID=">",
+            delay=0,
+            errDelay=1,
+            errCallback=logger.error,
     ):
         self.createQueue(queue, group=group, existOk=True)
         logger.debug("Subscribing Messages")
@@ -226,23 +258,23 @@ class MQ(object):
                 self.client.xack(queue, group, message["id"])
 
     def recvPendingMessagesInfo(
-        self, queue, group="default", consumer="unknown", start="-", end="+", count=None
+            self, queue, group="default", consumer="unknown", start="-", end="+", count=None
     ):
         return self.client.xpending_range(
             queue, group, start, end, count, consumername=consumer
         )
 
     def retryPendingMessages(
-        self,
-        queue,
-        group="default",
-        consumer="unknown",
-        count=100,
-        maxTimes=3,
-        timeout=1,
-        errCallback=logger.error,
-        maxlen=1000,
-        trimImmediately=False,
+            self,
+            queue,
+            group="default",
+            consumer="unknown",
+            count=100,
+            maxTimes=3,
+            timeout=1,
+            errCallback=logger.error,
+            maxlen=1000,
+            trimImmediately=False,
     ):
         logger.debug("Retrying Pending Messages")
 
